/*****************************************************************************
*  Copyright 2005 Alt-N Technologies, Ltd. 
*
*  Licensed under the Apache License, Version 2.0 (the "License"); 
*  you may not use this file except in compliance with the License. 
*  You may obtain a copy of the License at 
*
*      http://www.apache.org/licenses/LICENSE-2.0 
*
*  Unless required by applicable law or agreed to in writing, software 
*  distributed under the License is distributed on an "AS IS" BASIS, 
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
*  See the License for the specific language governing permissions and 
*  limitations under the License.
*****************************************************************************/

#include <sys/types.h>
#include <ctype.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/nameser.h>
#include <resolv.h>

#include <stdio.h>
#include <string.h>

#include "dkim.h"
#include "dns.h"



bool IsPolicyRecord( char* szTxt )
{
	// TODO: Replace with TXT record verification when available
	return ( szTxt[0] != '\0' );
}


bool IsKeyRecord( char* szTxt )
{
	// TODO: Replace with TXT record verification when available	
	return (szTxt[0] != '\0');
}

static unsigned short getshort(unsigned char *cp) {
  return (cp[0] << 8) | cp[1];
}

////////////////////////////////////////////////////////////////////////////////
// 
// _DNSGetPolicy for UNIX and win2k-
//
////////////////////////////////////////////////////////////////////////////////
int _DNSGetPolicy( const char* szSubDomain, char* Buffer, int nBufLen )
{
    u_char answer[PACKETSZ+1]; 
	int answerlen;
    int ancount, qdcount;	/* answer count and query count */
    u_char *eom, *cp;
    u_short type, rdlength;		/* fields of records returned */
	int rc;
	char* bufptr;

	// TODO: Query for a DKSP RR first

    answerlen = res_query( szSubDomain, C_IN, T_TXT, answer, sizeof(answer));

	if( answerlen < 0 )
	{
		if( h_errno == TRY_AGAIN )
			return DNSRESP_TEMP_FAIL;
		else
			return DNSRESP_PERM_FAIL;
	}

    qdcount = getshort( answer + 4); /* http://crynwr.com/rfc1035/rfc1035.html#4.1.1. */
    ancount = getshort( answer + 6);


    eom = answer + answerlen;
    cp  = answer + HFIXEDSZ;
	
	while( qdcount-- > 0 && cp < eom ) 
	{
		rc = dn_expand( answer, eom, cp, Buffer, nBufLen );
		if( rc < 0 ) {
			return DNSRESP_PERM_FAIL;
		}
		cp += rc + QFIXEDSZ;
    }
	
    while( ancount-- > 0 && cp < eom ) 
	{
		rc = dn_expand( answer, eom, cp, Buffer, nBufLen );
		if( rc < 0 ) {
			return DNSRESP_PERM_FAIL;
		}
		
		cp += rc;
		
		if (cp + RRFIXEDSZ >= eom) return DNSRESP_PERM_FAIL;
		
		type = getshort(cp + 0); /* http://crynwr.com/rfc1035/rfc1035.html#4.1.3. */
		rdlength = getshort(cp + 8);
		cp += RRFIXEDSZ;
		
		if( type != T_TXT ) {
			cp += rdlength;
			continue;
		}
		
		bufptr = Buffer;
		while (rdlength && cp < eom) 
		{
			int cnt;
			
			cnt = *cp++;		 /* http://crynwr.com/rfc1035/rfc1035.html#3.3.14. */
			if( bufptr-Buffer + cnt + 1 >= nBufLen )
				return DNSRESP_PERM_FAIL;
			if (cp + cnt > eom)
				return DNSRESP_PERM_FAIL;
			memcpy( bufptr, cp, cnt);
			rdlength -= cnt + 1;
			bufptr += cnt;
			cp += cnt;
			*bufptr = '\0';
		}
		
		return DNSRESP_SUCCESS;
    }

	return DNSRESP_PERM_FAIL;
}


////////////////////////////////////////////////////////////////////////////////
// 
// _DNSGetKey for UNIX and win2k-
//
////////////////////////////////////////////////////////////////////////////////
int _DNSGetKey( const char*szSubDomain, char* Buffer, int nBufLen )
{
    u_char answer[2*PACKETSZ+1]; 
	int answerlen;
    int ancount, qdcount;	/* answer count and query count */
    u_char *eom, *cp;
    u_short type, rdlength;		/* fields of records returned */
	int rc;
	char* bufptr;

	// TODO: Query for a DKK RR first


    answerlen = res_query( szSubDomain, C_IN, T_TXT, answer, sizeof(answer));

	if( answerlen < 0 )
	{
		if( h_errno == TRY_AGAIN )
			return DNSRESP_TEMP_FAIL;
		else
			return DNSRESP_PERM_FAIL;
	}

    qdcount = getshort( answer + 4); /* http://crynwr.com/rfc1035/rfc1035.html#4.1.1. */
    ancount = getshort( answer + 6);


    eom = answer + answerlen;
    cp  = answer + HFIXEDSZ;
	
	while( qdcount-- > 0 && cp < eom ) 
	{
		rc = dn_expand( answer, eom, cp, Buffer, nBufLen );
		if( rc < 0 ) {
			return DNSRESP_PERM_FAIL;
		}
		cp += rc + QFIXEDSZ;
    }
	
    while( ancount-- > 0 && cp < eom ) 
	{
		rc = dn_expand( answer, eom, cp, Buffer, nBufLen );
		if( rc < 0 ) {
			return DNSRESP_PERM_FAIL;
		}
		
		cp += rc;
		
		if (cp + RRFIXEDSZ >= eom) return DNSRESP_PERM_FAIL;
		
		type = getshort(cp + 0); /* http://crynwr.com/rfc1035/rfc1035.html#4.1.3. */
		rdlength = getshort(cp + 8);
		cp += RRFIXEDSZ;
		
		if( type != T_TXT ) {
			cp += rdlength;
			continue;
		}
		
		bufptr = Buffer;
		while (rdlength && cp < eom) 
		{
			int cnt;
			
			cnt = *cp++;		 /* http://crynwr.com/rfc1035/rfc1035.html#3.3.14. */
			if( bufptr-Buffer + cnt + 1 >= nBufLen )
				return DNSRESP_PERM_FAIL;
			if (cp + cnt > eom)
				return DNSRESP_PERM_FAIL;
			memcpy( bufptr, cp, cnt);
			rdlength -= cnt + 1;
			bufptr += cnt;
			cp += cnt;
			*bufptr = '\0';
		}
		
		return DNSRESP_SUCCESS;
    }

	return DNSRESP_PERM_FAIL;
}


////////////////////////////////////////////////////////////////////////////////
// 
// DNSGetPolicy
//
// Pass in the user and domain. This function combines them into the FQDN
// to query
//
////////////////////////////////////////////////////////////////////////////////
int DNSGetPolicy( const char* szUser, const char* szDomain, char* Buffer, int nBufLen )
{
	char szSubDomain[MAX_DOMAIN+1];
	int nUserStrLen = 0;

	if( szUser && szUser[0] )
	{
		nUserStrLen = strlen( szUser );
	}


	if( strlen(szDomain) + nUserStrLen > MAX_DOMAIN - 20 )
		return DNSRESP_DOMAIN_NAME_TOO_LONG;

	// construct the subdomain to query

	if( nUserStrLen > 0 )
	{
		strcpy( szSubDomain, szUser );
		strcat( szSubDomain, "._policy._domainkey." );
	}
	else
	{
		strcpy( szSubDomain, "_policy._domainkey." );
	}
	strcat( szSubDomain, szDomain );

	// Initialize out parameter
	Buffer[0] = '\0';

	return _DNSGetPolicy( szSubDomain, Buffer, nBufLen );
}

////////////////////////////////////////////////////////////////////////////////
// 
// DNSGetPolicy
//
// Pass in the FQDN to get the policy
//
////////////////////////////////////////////////////////////////////////////////
int DNSGetPolicy( const char* szFQDN, char* Buffer, int nBufLen )
{
	if (strlen(szFQDN) > MAX_DOMAIN)
		return DNSRESP_DOMAIN_NAME_TOO_LONG;

	// Initialize out parameter
	Buffer[0] = '\0';

	return _DNSGetPolicy( szFQDN, Buffer, nBufLen );
}


////////////////////////////////////////////////////////////////////////////////
// 
// DNSGetKey
//
// Pass in the selector name and domain name. This function combines them into 
// the fqdn to query.
//
////////////////////////////////////////////////////////////////////////////////
int DNSGetKey( const char* szSelector, const char* szDomain, char* Buffer, int nBufLen )
{
	char szSubDomain[MAX_DOMAIN+1];
	int nSelectorStrLen = 0;

	if( szSelector && szSelector[0] )
	{
		nSelectorStrLen = strlen( szSelector );
	}

	if( strlen(szDomain) + nSelectorStrLen > MAX_DOMAIN - 11 )
		return DNSRESP_DOMAIN_NAME_TOO_LONG;

	// construct the subdomain to query

	if( nSelectorStrLen > 0 )
	{
		strcpy( szSubDomain, szSelector );
		strcat( szSubDomain, "._domainkey." );
	}
	else
	{
		strcpy( szSubDomain, "_domainkey." );
	}

	strcat( szSubDomain, szDomain );

	// Initialize out parameter
	Buffer[0] = '\0';

	return _DNSGetKey( szSubDomain, Buffer, nBufLen );
}


////////////////////////////////////////////////////////////////////////////////
// 
// DNSGetKey
//
// Pass in the FQDN to get the selector
//
////////////////////////////////////////////////////////////////////////////////
int DNSGetKey( const char* szFQDN, char* Buffer, int nBufLen )
{
	if (strlen(szFQDN) > MAX_DOMAIN)
		return DNSRESP_DOMAIN_NAME_TOO_LONG;

	// Initialize out parameter
	Buffer[0] = '\0';

	return _DNSGetKey( szFQDN, Buffer, nBufLen );
}
